/* XLib graphics driver capable of MITSHM
 * Also handles 16 and 24bpp
 *
 * by Jay Kominek <jkominek@debian.org>
 *
 * Based on the plain ole Xlib graphics driver by:
 * Johannes Hirche
 */

#include "dlinux.h"

#ifdef HAVE_X11
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include "error.h"
#include "gr.h"
#include "mono.h"
#include "key.h"
#include "timer.h"
#include "args.h"

/* Here are the keycodes for X->descent */
#include "xlib_keycode_to_descent.h"
#include "xlib_special_to_descent.h"

/* Prototype it so that x11_setpalette() can call it */
void x11_update(void);

/* Locals */
int granularity=0;

/* Xlib Variables */
Display *display = NULL;
Screen *screen;
Window window;
Colormap cmap = 0;
GC gc;
XImage *image = NULL;
unsigned long pen_fg,pen_bg;
XEvent event;
XVisualInfo visualinfo;
unsigned long r_mask, g_mask, b_mask;
/* SHM Stuff */
XShmSegmentInfo shminfo;
Bool usingShm;

/* size of window */
int xsize = 0, ysize = 0;

/* current colormap */
XColor colors[256];

/* Size of one frame-buffer (xsize*ysize) */
unsigned long bytes_per_page=0;

/* Pointer to draw-buffer */
unsigned char *gr_mem=NULL;

/* initialized? */
int running=0;

/* array of pointers to the different pages */
unsigned char **gr_mem_pages;

/* Current bit depth */
int bpp;

/* Graphics shtuff */
int x11_set_mode(int mode) {
  XSizeHints hints;
  XWMHints wmhints;
  XClassHint classhints;
  XTextProperty name;
  XSetWindowAttributes attributes;
  XGCValues gc_val;
  unsigned long gc_valmask;
  
  int i, major, minor;
  Bool shared;
  char *namestring[] = {"Descent"};
  
  if(running) return 1; /* no mode switch supported yet! */
  
  /* boy is this ugly... */
  switch(mode){
  case 0:
  case 1: xsize = 320; ysize = 200; break;
  case 2: xsize = 320; ysize = 240; break;
  case 3: xsize = 360; ysize = 200; break;
  case 4: xsize = 360; ysize = 240; break;
  case 5: xsize = 376; ysize = 282; break;
  case 6: xsize = 320; ysize = 400; break;
  case 7: xsize = 320; ysize = 480; break;
  case 8: xsize = 360; ysize = 400; break;
  case 9: xsize = 360; ysize = 480; break;
    /* FIXME add the other modes... Got tired of typing the screensizes */
  default: xsize = 320; ysize = 200;
  }
  granularity=32;
  running=1;

  /* This will keep all well behaved window managers from letting the user
     alter the size of our window */
  hints.min_width  = hints.max_width  = hints.base_width  = xsize;
  hints.min_height = hints.max_height = hints.base_height = ysize;
  
  /* Fix this: it only uses $DISPLAY to get the display */
  if(!(display = XOpenDisplay(NULL))) {
    return(1);
  }

  if((usingShm=XShmQueryExtension(display))) {
    XShmQueryVersion(display,&major,&minor,&shared);
    printf("\nUsing SHM extension. Version %i.%i %s shared pixmaps",
	   major, minor,
	   shared ? "with" : "without");
  } else {
    printf("\nSHM extension not present.");
  }
  
  /* Get the default screen */
  screen = DefaultScreenOfDisplay(display);
  /* Get the depth of the screen */
  bpp = DefaultDepthOfScreen(screen);

  /* Get an appropriate visual */
  if(bpp==8) {
    if(!XMatchVisualInfo(display, DefaultScreen(display),
			 bpp, PseudoColor, &visualinfo)) {
      printf("Could not get PseudoColor visual!\n");
      return(1);
    }
  } else {
    if(!XMatchVisualInfo(display, DefaultScreen(display),
			 bpp, TrueColor, &visualinfo))
      printf("Could not get TrueColor visual!\n");
  }

  if(bpp==8) {
    cmap = XCreateColormap(display, RootWindowOfScreen(screen),
			   visualinfo.visual, AllocAll);
  } else cmap = DefaultColormapOfScreen(screen);

  XStringListToTextProperty(namestring,1,&name);

  attributes.colormap = cmap;
  
  /* Create the window */
  window = XCreateWindow(display,                     /* Display */
			 RootWindowOfScreen(screen),  /* Parent Window */
			 0,0,                         /* Suggested location on screen */
			 xsize, ysize,                /* Size of window */
			 4,                           /* Width of borders */
			 bpp,                         /* Screen depth */
			 CopyFromParent,              /* Class */
			 visualinfo.visual,           /* Visual */
			 CWColormap,                  /* Window attribute mask */
			 &attributes);                /* Window attributes */

  if(!window) {
    /* Wow, what kind of broken system does this guy have? */
    printf("Could not open a window!\n");
    return(1);
  }

  /* Create the graphics context */
  gc_val.graphics_exposures = False;
  gc_val.foreground         = pen_fg;
  gc_val.background         = pen_bg;
  gc_valmask                = GCGraphicsExposures | GCForeground | GCBackground;
  gc = XCreateGC(display, window, gc_valmask, &gc_val);

  /* Setup hints */
  hints.flags = PSize | PMinSize | PMaxSize;
  wmhints.input = True;
  wmhints.flags = InputHint;
  classhints.res_class = classhints.res_name  = "DESCENT";

  /* Send information to the window manager. (If we actually had argc and argv, this is
     what would make us dockable.) */
  XSetWMProperties(display, window, &name, &name, Args, Num_args, &hints,
		   &wmhints, &classhints);

  /* Set out event mask */
  XSelectInput(display, window, ExposureMask | ButtonPressMask | ButtonReleaseMask |
	       ButtonMotionMask | KeyPressMask | KeyReleaseMask| StructureNotifyMask);

  XMapRaised(display, window);
  XClearWindow(display, window);
  XFlush(display);

  /* Wait for window to be mapped */
  do
    XNextEvent(display,&event);
  while(event.type!=Expose);

  if(usingShm) {
    image = XShmCreateImage(display, visualinfo.visual, bpp, ZPixmap, NULL,
			    &shminfo, xsize, ysize);
    if(!image) {
      printf("Failed to get SHM image\n");
      return(1);
    }
  } else {
    image = XGetImage(display,window,0,0,xsize,ysize,AllPlanes,ZPixmap);
    if(!image) {
      printf("Failed to get image\n");
      return(1);
    }
  }

  /* Get the masks here, since they won't change later */
  r_mask = image->red_mask;
  g_mask = image->green_mask;
  b_mask = image->blue_mask;

  /* Get SHM information from the image */
  if(usingShm) {
    shminfo.shmid        = shmget(IPC_PRIVATE, image->bytes_per_line*image->height,
				  IPC_CREAT|0777);
    shminfo.shmaddr      = shmat(shminfo.shmid, 0, 0);
    image->data          = shminfo.shmaddr;
    if(!image->data) {
      XDestroyImage(image);
      printf("Could not allocate shared memory\n");
      return(1);
    }

    shminfo.readOnly = False; /* We have to be able to alter the contents of the memory */
    if(!XShmAttach(display,&shminfo)) {
      XDestroyImage(image);
      printf("Could not attach shared memory to the X server\n");
      return(1);
    }
  }

  /* Don't give the rest of the Descent code the real location of
     memory, since the real location is bigger for 16 and 24bpp */
  gr_mem = malloc(xsize*ysize);
  if(!gr_mem) {
    printf("Failed to malloc() memory for the graphics buffer\n");
    return(1);
  }
  gr_mem_pages = malloc(sizeof(unsigned char *)*8);
  for(i = 0; i < 8; i++)
    gr_mem_pages[i] = NULL;

  if(bpp==8)
    for(i=0;i<256;i++)
      colors[i].flags=DoRed|DoGreen|DoBlue;
  /* Should check for modeinfo here. (?) */

  return 0;
}

void x11_set_palette(char *pal) {
  register int i;

  if (!running||!colors) return;

  if(bpp==8)
    {
      for(i = 0; i < 256; i++)
	{
	  colors[i].pixel = i;
	  colors[i].red   = ((*pal++)<<10);
	  colors[i].green = ((*pal++)<<10);
	  colors[i].blue  = ((*pal++)<<10);
	}
      XStoreColors(display,cmap,colors,256);
    }
  else
    {
      for(i = 0; i < 256; i++)
	{
	  colors[i].red   = ((*pal++)<<10);
	  colors[i].green = ((*pal++)<<10);
	  colors[i].blue  = ((*pal++)<<10);
	}
      x11_update();
    }
  return;
}

void x11_get_palette(char *pal) {
  register int i;
  if (!running||!colors) return;
  
  for(i=0; i<256; i++) {
    *pal++ = (colors[i].red)   >> 10;
    *pal++ = (colors[i].green) >> 10;
    *pal++ = (colors[i].blue)  >> 10;
  }
  return;
}

void x11_set_drawing_page(int page) {
  return;
}

void x11_set_viewing_page(int page) {
  return;
}

void x11_bank(int bank) {
}

unsigned char *x11_get_buffer() {
  if(!running)return NULL;
  else
    return gr_mem;
}

int x11_get_gran() {
  return granularity;
}

/* Here we copy the memory from gr_mem into image->data, but convert it
   to the appropriate bit depth as we go. If code to enlarge the window
   is ever added, it should go in here, before the 8bpp to Xbpp code. */ 
void x11_update(void) {
  register unsigned int wh = image->bytes_per_line/(bpp>>3)*image->height;
  register unsigned char *buffer;

  switch(bpp) {
  case 8:
    /* 8bpp data is just fine and dandy, its the right size and has
       the right values in it */
    memcpy(image->data,gr_mem,image->bytes_per_line*image->height);
    break;
  case 16:
    /* 16 and 24 bpp code sort of by Brad Hughes <bhughes@arn.net> */
    {
      register unsigned short *im = (unsigned short *)image->data;
      buffer = gr_mem;
      while(--wh)
	*(im++) = (colors[*buffer].red & r_mask) |
	  ((colors[*buffer].green>>5) & g_mask) |
	  ((colors[*(buffer++)].blue>>11) & b_mask);
      *(im++) = (colors[*buffer].red & r_mask) |
	((colors[*buffer].green>>5) & g_mask) |
	((colors[*buffer].blue>>11) & b_mask);
    }
    break;
  case 24:
    {
      unsigned char *im = (unsigned char *)image->data;
      buffer = gr_mem;
      while(--wh) {
	*(im++) = (unsigned char)(colors[*buffer].blue>>8);
	*(im++) = (unsigned char)(colors[*buffer].green>>8);
	*(im++) = (unsigned char)(colors[*(buffer++)].red>>8);
      }
      *(im++) = (unsigned char)(colors[*buffer].blue>>8);
      *(im++) = (unsigned char)(colors[*buffer].green>>8);
      *(im)   = (unsigned char)(colors[*(buffer++)].red>>8);
    }
    break;
  case 32:
    /* I still can't get proper information on this bit depth so as
       to support it. */
    break;
  default:
    /* <8bpp can't be done, and I'd need xdpyinfo output from a 15bpp
       display to support it. */
    printf("Your bit depth is not supported!\n");
    break;
  }

  /* Send out the image */
  if(usingShm)
    XShmPutImage(display,window,gc,image,0,0,0,0,xsize,ysize,False);
  else
    XPutImage(display,window,gc,image,0,0,0,0,xsize,ysize);
  /* Flush it to the server */
  XFlush(display);
}

inline int keysym_to_descent(KeySym keysym)
{
  if(keysym<255)
    /* no special key */
      return xlib_keycode_to_descent[keysym];
  return xlib_special_to_descent[keysym&0x00ff];
}

int x11_bm_type(void){
  return BM_LINEAR;
}

void x11_modex_plane(int plane){
  /* FIXME What is this ?? */
}

int x11_keyboard_init(void) {
  return 0;
}

void x11_keyboard_update(void) {
  int kcode,temp;
  KeySym keysym;
  
  if(!XPending(display))
    return;
  
  XNextEvent(display,&event);
  switch(event.type){
  case KeyPress:
    keysym = XLookupKeysym(&event.xkey,0);
    kcode  = keysym_to_descent(keysym);
    keyd_last_pressed = kcode;
    keyd_time_when_last_pressed = timer_get_fixed_secondsX();
    if(!keyd_pressed[kcode])
      {
	key_data.NumDowns[kcode]++;
	keyd_pressed[kcode] = 1;
	key_data.TimeKeyWentDown[kcode] = timer_get_fixed_secondsX();	   
      }
    kcode = kcode | (keyd_pressed[KEY_LSHIFT] || keyd_pressed[KEY_RSHIFT] ? KEY_SHIFTED : 0)
      | (keyd_pressed[KEY_LALT]  || keyd_pressed[KEY_RALT]  ?KEY_ALTED  : 0)
      | (keyd_pressed[KEY_LCTRL] || keyd_pressed[KEY_RCTRL] ?KEY_CTRLED : 0);
    if((temp = key_data.keytail + 1) >= KEY_BUFFER_SIZE)
      temp = 0;
    if(temp!=key_data.keyhead)
      {
	key_data.keybuffer[key_data.keytail] = kcode;
	key_data.time_pressed[key_data.keytail] = keyd_time_when_last_pressed;
	key_data.keytail = temp;
      }
    break;
  case KeyRelease:
    keysym = XLookupKeysym(&event.xkey,0);
    kcode  = keysym_to_descent(keysym);
    keyd_last_released  = kcode;
    keyd_pressed[kcode] = 0;
    key_data.NumUps[kcode]++;
    key_data.TimeKeyHeldDown[kcode] +=
      timer_get_fixed_secondsX() - key_data.TimeKeyWentDown[kcode];
    break;
  }
}

void x11_keyboard_close(void) {
  return;
}

/* Mouse stuff should be down here. */
int x11_mouse_init(void) {
  return -1;
}

void x11_mouse_close(void) {
  return;
}

void x11_mouse_update(int *dx, int *dy, int *buttons) {
        return;
}

#endif
